/* -------------------------------------------------------------------	*/
/* Copyright (C) 2008 Robert R. Bruner.  All rights reserved.		*/
/* Contact the author at						*/
/*                    rrb@math.wayne.edu				*/
/*                    Dept of Mathematics				*/
/*                    Wayne State University				*/
/*                    Detroit, Michigan  48202				*/
/*                    USA						*/
/*									*/
/* Version 1.8								*/
/* -------------------------------------------------------------------- */



#include <stdio.h>
#include <stdlib.h>
/* Try commenting this out
    #include <malloc.h> */

#include <util.h>
#include <steen.h>
#include <structur.h>
#include <storage.h>
#include <io.h>
#include <errors.h>

typedef struct Steen_Op {
                    int first_sq;
        struct Steen_Op *first_rest;
        struct Steen_Op *other_terms; } Steen_Op;

typedef struct module_element {
                               int first;
             struct module_element *next;} module_element;

module_element ***sq_value=NULL;
int numgen, *degrees;
int initialized=0;

/* ------------------------------------------------------------------ */
module_element * copy_module_element(x) module_element *x;
{ module_element *z;

if (x == NULL) return(NULL);

z = (module_element *) malloc(sizeof(module_element));
z->first = x->first;
z->next = copy_module_element(x->next);
return(z);
}

/* ------------------------------------------------------------------ */
module_element * insert_gen(g,e)  int g; module_element *e;
{
  if ((e==NULL) || g < e->first)
  {
    module_element *e1=(module_element *)malloc(sizeof(module_element));

    e1->first=g;
    e1->next=e;
    return(e1);
  }
  if (g == e->first)  /*bad data file */
  {
    fprintf(stderr,"bad data file: Def\n");
    return (e->next);
  }
  e->next=insert_gen(g,e->next);
  return(e);
}

/* ------------------------------------------------------------------ */
void initialize()
{ FILE *inputfile;
  int i, gen, op, count, g;

  if (NULL == (inputfile=fopen("Def","r")))
    terminate("\nCannot open module definition: Def\n"
              ,BAD_FILE);

  fscanf(inputfile, "%d", &numgen);

  degrees=(int *)malloc(numgen * sizeof(int));
  for (gen=0;gen<numgen;gen++)
    fscanf(inputfile, "%d ",&degrees[gen]);

  sq_value=(module_element ***)malloc(numgen*sizeof(module_element **));
  for (gen=0;gen<numgen;gen++)
  {
    sq_value[gen]=(module_element **)malloc(
         (1+degrees[numgen-1] - degrees[gen])*sizeof(module_element *));
    sq_value[gen][0] = (module_element *)malloc(sizeof(module_element));
    sq_value[gen][0]->first = gen;
    sq_value[gen][0]->next = NULL;
    for(i=1;i<=degrees[numgen-1] - degrees[gen];i++)
      sq_value[gen][i]=NULL;
  }
                /* sq_value[gen][op]'s are initially NULL */

  while (EOF != fscanf(inputfile, "%d %d %d", &gen,&op,&count))
    for(i=0;i<count;i++)
    { fscanf(inputfile, "%d", &g);
      sq_value[gen][op]=insert_gen(g,sq_value[gen][op]);
    }
fclose(inputfile);
initialized=1;
return;
}

/* ------------------------------------------------------------------ */
int mod_basis(t,adegree) int t, **adegree;
{ int i,count;

if (!initialized) initialize();

for (count = i = 0; i < numgen; i++)
  if (degrees[i] <= t) count++;

*adegree = degrees;
return(count);
}

/* ------------------------------------------------------------------ */
void print_module_element(e)  module_element *e;
{
  module_element *i=NULL;

  if (e != NULL)
    for(i=e;i != NULL;i=i->next)
      printf("%d + ",i->first);
  printf("done\n");
  return;
}

/* --------------------------------------------------------------- */
module_element * modadd(x,y) module_element *x,*y;
{
  module_element *x1=NULL,*y1=NULL;
                                   /* destroys x and y as it adds */
  if (x==NULL) return(y);
  if (y==NULL) return(x);
  if (x->first < y->first)
  {
    x->next = modadd(x->next,y);
    return(x);
  }
  if (x->first > y->first)
  {
    y->next = modadd(x,y->next);
    return(y);
  }
  /* first terms are equal, hence cancel */
  x1=x->next;
  y1=y->next;
  free(x);
  free(y);
  return(modadd(x1,y1));
}

/* --------------------------------------------------------------- */
module_element * actone(op,x) int op; module_element *x;
{
  module_element *e=NULL,*f=NULL;
  char errmess[100];

  if (x==NULL)
    return(x);
  if ((x->first >= numgen) || (x->first < 0))
  { sprintf(errmess,"\nBad gen %d out of range %d to %d in actone\n",
            x->first,0,numgen-1);
    terminate(errmess,MODERR);
  }
  if (op + degrees[x->first] > degrees[numgen-1])
    e=NULL;
  else
    e = copy_module_element(sq_value[x->first][op]);
  f=actone(op,x->next);  /* if op is acting on an element x which is the
                            sum of more than one element, say
                            x1 + x2 + ... + xn, then f will equal op
                            acting on x2 + ... + xn                       */
  return(modadd(e,f));
}

/* ------------------------------------------------------------------ */
module_element * act(tree,x) Steen_Op *tree; module_element *x;
{
  module_element *e=NULL,*f=NULL;
  char errmess[100];

  if ((x->first >= numgen) || (x->first < 0))
  { sprintf(errmess,"\nBad gen %d out of range %d to %d in act\n",
            x->first,0,numgen-1);
    terminate(errmess,MODERR);
  }
  if (tree == NULL)
    return(x);
  if ((tree->first_rest) == NULL)   /* tree->first_sq in modadd statement
                                       below will act on x                */
    e=x;
  else
    e=act(tree->first_rest,x);
  if ((tree->other_terms) == NULL)  /* add NULL to modadd statement below */
    f=NULL;
  else
    f=act(tree->other_terms,x);
  return(modadd(actone(tree->first_sq,e),f));
}

/* ------------------------------------------------------------------  */
Steen_Op * add_Steen_Op(y,z) Steen_Op *y, *z;
{
if (y == NULL) return(z);
else if (z == NULL) return(y);
else if (y-> first_sq < z->first_sq)
{ y->other_terms = add_Steen_Op(y->other_terms,z);
  return(y);
}
else if (y-> first_sq > z->first_sq)
{ z->other_terms = add_Steen_Op(z->other_terms,y);
  return(z);
}
else /* first terms equal */
{ Steen_Op *w = add_Steen_Op(y->first_rest,z->first_rest);
  if (w != NULL) { y->other_terms =
                       add_Steen_Op(y->other_terms,z->other_terms);
                   free(z);
                   y->first_rest = w;
                   return(y);
                 }
  else                 /* first terms cancelled     */
  w = add_Steen_Op(y->other_terms,z->other_terms);
  free(y); free(z);
  return(w);
}
}

/* ------------------------------------------------------------------   */
int last_op_in(x,Op) TERM *x; Internal_Op Op;
                         /* return the sequence number of the last op   */
                         /* in x, and set Op to that operation          */
{
int bit,disp,i;

if (x == NULL) return(-1);
if (first_op(Op,x->deg) < 0) return(-1);
for (i=((x->len-1) >> 3); (i >=0)  && (x->op[i] == 0); i--);
if (i < 0) return(-1);
for (disp=7,bit=1; !(x->op[i] & bit); disp--, bit <<=1);
i = (i << 3) + disp;
adv_op(Op,i);
return(i);
}

/* ------------------------------------------------------------------ */
Steen_Op * expand_aux(x) TERM *x;
{ /* expand returns a sum y of products of Sq(i)'s which equals the
     sum of Milnor basis elements contained in the TERM * x             */
  /* Cannibalizes x.  (x is undefined upon exit)                        */

Steen_Op *y1, *y = NULL;
Internal_Op RR,R,S;
int bit, seq, last, r, deg, i;
TERM *t;

for (seq = last_op_in(x,R); seq >=0; seq = last_op_in(x,R))
{  for (last=N_ALG; R[--last] == 0; );
   if (last == 0)
   {  y1 = (Steen_Op *) malloc(sizeof(Steen_Op));
      y1->first_sq = R[0];
      y1->first_rest = y1->other_terms = NULL;
      y = add_Steen_Op(y,y1);
      x->op[0] = 0;
   }
   else                  /* last > 0   */
   {  r = R[last];
      for (i=0; i < N_ALG; i++) RR[i] = R[i];
      RR[last] = 0;
      RR[last-1] += r;
      r <<= last;
                                 /* add Sq(r) * expand(Sq(RR)) to y,
                                    subtract Sq(r)*Sq(RR) from x     */
      deg = degree(RR);
      t = get_TERM(dimA(deg));
      t->gen = 0;
      t->deg = deg;
      zero_op(t);
                                /* set bit seqno(RR) of t to 1      */
      bit = Seqno(RR);
      t->op[bit >> 3] = (0x80) >> (bit & 7);

      y1 = (Steen_Op *) malloc(sizeof(Steen_Op));
      y1->first_sq = r;
      y1->first_rest = expand_aux(t);
      return_TERM(t);
      y1->other_terms = NULL;
      y = add_Steen_Op(y,y1);

      S[0] = r;
      for (i=1; i < N_ALG; i++) S[i] = 0;
      mult(S,RR,x);
   }
}
return(y);
}

/* ------------------------------------------------------------------ */
Steen_Op * expand(op) Internal_Op op;
{
  Steen_Op *s=NULL;
  Element *e;
  TERM *t;
  int deg;
  unsigned seq;

  deg=degree(op);
  if (deg == 0)
    return(s);
  t = get_TERM(dimA(deg));
  t->deg = deg;
  t->gen = 0;
  zero_op(t);
  seq=Seqno(op);
  t->op[seq >> 3] = (0x80) >> (seq & 7);
  return(expand_aux(t));
}

/* ------------------------------------------------------------------ */
void modact(dim,dx,t,Op,opseqno,dg,g_degree)
       int dim,t,opseqno,g_degree; TERM **dx, **dg; Internal_Op Op;
{
int g;
module_element *value, *newvalue;

if (!initialized) initialize();
for (value=NULL,g=dim; g-- > 0;)
  if ((dg[g] != NULL) && is_not_zero(dg[g]))
    value = insert_gen(g,value);

newvalue = act(expand(Op),value);

for (g=0; g < dim; g++)
if ((newvalue == NULL) || (g < newvalue->first)) zero_op(dx[g])
else { dx[g]->op[0] = 0x80;
       newvalue = newvalue->next;
     }
return;
}

