/* -------------------------------------------------------------------	*/
/* Copyright (C) 2000 Robert R. Bruner.  All rights reserved.		*/
/* Contact the author at						*/
/*                    rrb@math.wayne.edu				*/
/*                    Dept of Mathematics				*/
/*                    Wayne State University				*/
/*                    Detroit, Michigan  48202				*/
/*                    USA						*/
/*									*/
/* Version 1.6								*/
/* -------------------------------------------------------------------- */


#include <stdio.h>
#include <stdlib.h>

#include <errors.h>
#include <storage.h>
#include <io.h>
#include <steen.h>
#include <module.h>
#include <structur.h>
#include <util.h>

char Status_fmt[] = "%10d";   /* format of status in files Im and Ker */
char Diff_hdr[] = "%10d %10d";   /* format of header in Diff files    */


main(argc,argv) int argc; char *argv[];
{
int i,j,nread,numops,deg,dim,found,status,g,gen,op,
    lgen,lop,                /* leading gen and op                     */
    s,t,                     /* bidegree to do                         */
    degree_done,             /* of Diff.s-1 then of Diff.s             */
    dim_codomain,            /* dim over A of codomain                 */
    image_dim,               /* dim over A of codomain in degree < t   */
    **image_gen,**image_op,  /* (gen,op) for corresp image entry       */
    *degree;                 /* internal degree of gens of C.s-1       */
Element *e;                  /* image elements                         */
/* ------------------------------------------------------------------- */

Element  ***image_x,
         ***image_dx;
                             /* These are each arrays of arrays of     */
                             /* pointers to Elements, such that the    */
                             /* non-NULL image_dx[g][o] are a basis    */
                             /* for the image, and                     */
                             /*  d(image_x[g][o]) = image_dx[g][o]     */
                             /* Further, image_dx[g][o] = o * g +      */
                             /*  higher terms                          */

/* ------------------------------------------------------------------  */
FILE *Diff,                  /* Diff.s-1                               */
     *Im,                    /* Im_s.t                                 */
     *OIm;                   /* OIm_s.t                                */
char filename[20],message[100];


/* -------------- check calling sequence -----------------------------  */
if (argc != 3) {
  printf("\nWrong number %d of args  Should be 2.\n",argc-1);
  for (i=0;i<argc;i++) fprintf(stderr,"%3d: %s\n",i,argv[i]);
  exit(1);
 }
/* ------------------------------ get s,t ------------------------------ */

sscanf(argv[1],"%d",&s);
sscanf(argv[2],"%d",&t);

/* ----------- process Diff.s-1   -------------------------------------  */
/* Check previous filtration (or module) for completness, get dimension  */
/*  and degrees of gens to compute A dim of the codomain                 */

if (s > 0) 
{  sprintf(filename,"Diff.%d",s-1);
   if ((Diff = fopen(filename,"r")) == NULL) 
    { sprintf(message,"\nFile not found: %s\n",filename);
      terminate(message,BAD_DIFF);
     }

   if ((nread = fscanf(Diff,Diff_hdr,&dim_codomain,&degree_done)) != 2) 
    { sprintf(message,"\nBad %s: read only %d variables from header.",
                     filename,nread);
      terminate(message,BAD_DIFF);
     }

   if (degree_done < t) 
    { sprintf(message,"\n%s incomplete.  Done through degree %d",
                     filename,degree_done);
      terminate(message,BAD_DIFF);
     }
/* read Diff.s-1 to get degrees of generators so we'll know how much   */
/*  space to allocate                                                  */

   degree = (int *)malloc(dim_codomain*sizeof(int));
   image_dim = 0;
   for (g=0; g<dim_codomain; g++) {
      fscanf(Diff,"%d",&degree[g]);
      skip_Element(Diff);
      if (degree[g] <= t) image_dim = 1 + g;
   }

   fclose(Diff);
                  /* done:  have dim_codomain, image_dim, and degrees */
}
                  /* s = 0   -    get these from the module instead   */
                  /*     (except dim_codomain, which is not needed)   */
else
   image_dim = mod_basis(t,&degree);

image_x = (Element ***) malloc(image_dim * sizeof(Element **));
image_dx = (Element ***) malloc(image_dim * sizeof(Element **));
image_gen = (int **) malloc(image_dim * sizeof(int *));
image_op = (int **) malloc(image_dim * sizeof(int *));

/* allocate entries in image and set them to NULL                      */

for (g=0; g<image_dim; g++)
{  deg = t - degree[g];
   numops = dimA(deg);
                   
   if (numops )
   { 
     image_x[g] = (Element **)malloc(numops * sizeof(Element *));
     image_dx[g] = (Element **)malloc(numops * sizeof(Element *));
     image_gen[g] = (int *) malloc (numops * sizeof(int));
     image_op[g] = (int *) malloc (numops * sizeof(int));
     for (i=0; i<numops; i++) 
        image_x[g][i] = image_dx[g][i] = NULL;
   }
   else    /* numops == 0  */
     image_dx[g] = NULL;
}

/* Now read Im and file it by leadgen and leadop, then write to      */
/*  OIm in increasing order                                          */

sprintf(filename,"Im_%d.%d",s,t);
if ((Im = fopen(filename,"r")) == NULL) {
     sprintf(message,"\nError opening %s ",filename);
     terminate(message,BAD_IM);
}
fscanf(Im,Status_fmt,&status);
if ((status != 1) && (status != 2))
{    sprintf(message,"\n%s is incorrect: status = %d\n",filename,status);
     terminate(message,BAD_IM);
}

while (fscanf(Im,"%d %d",&gen,&op) != EOF)
{ e = read_Element(Im);
  lgen = e->item[0]->gen;
  lop = first_seqno_in(e->item[0]);
  image_dx[lgen][lop] = e;
  image_x[lgen][lop] = read_Element(Im);
  image_gen[lgen][lop] = gen;
  image_op[lgen][lop] = op;
}

fclose(Im);


sprintf(filename,"OIm_%d.%d",s,t);
if ((OIm = fopen(filename,"w")) == NULL) {
     sprintf(message,"\nError opening %s ",filename);
     terminate(message,BAD_IM);
}

/* Scan image and write it in order of leading terms                   */

for (lgen=0; lgen<image_dim; lgen++)
  if (image_dx[lgen] != NULL)
    for (lop=0,numops = dimA(t-degree[lgen]);
	  lop < numops; lop++)
      if (image_dx[lgen][lop] != NULL)
       { fprintf(OIm,"%d %d %d %d\n",
	   lgen,lop,image_gen[lgen][lop],image_op[lgen][lop]);
	 write_Element(OIm,image_dx[lgen][lop],'c');
	 write_Element(OIm,image_x[lgen][lop],'c');
	 fprintf(OIm,"\n");
       }


fclose(OIm);
	


}
